from .common import UnmappedError, ResponseFormatError


def _construct_mapping():
    from .common import CODES as COMMON_CODES
    from .klass import CODES as KLASS_CODES
    from .lab import CODES as LAB_CODES
    from .pod import CODES as POD_CODES
    from .reservation import CODES as RESERVATION_CODES
    from .user import CODES as USER_CODES
    from .vm import CODES as VM_CODES

    result = {}
    for code in COMMON_CODES + KLASS_CODES + LAB_CODES + POD_CODES + RESERVATION_CODES + USER_CODES + VM_CODES:
        assert code.str_error not in result, code.str_error
        result[code.str_error] = code

    return result


NAME_CODE_MAPPING = _construct_mapping()


def error_decode(e):
    if not isinstance(e, dict):
        return ResponseFormatError()

    error_code = e['message']
    try:
        error_data = e['data']['message']
    except KeyError:
        error_data = None

    if error_code in NAME_CODE_MAPPING:
        return NAME_CODE_MAPPING[error_code](error_data)
    else:
        return UnmappedError(error_code, error_data)



import json
import socket
import sys

import click
from ..client import Client
from ..errors.common import AuthenticationError, InvalidConfig
from tabulate import tabulate


def client(system='default', config=None, config_path=None, catch_exceptions=True):
    if not catch_exceptions:
        return Client(system, config, config_path)

    try:
        netlab_client = Client(system, config, config_path)
    except InvalidConfig as e:
        click.secho("{}: {}".format(type(e).__name__, e), fg='red')
        click.secho("Check your config.json file for errors.")
        sys.exit(1)
    except json.JSONDecodeError as e:
        click.secho("{}: {}".format(type(e).__name__, e), fg='red')
        click.secho("Your config.json file has syntax errors and cannot be decoded.")
        sys.exit(1)
    except AuthenticationError as e:
        click.secho("{}: {}".format(type(e).__name__, e), fg='red')
        click.secho("Check your API key for problems. Make sure you are using the correct user, token, and source "
                    "IP address. Make sure your API key is active.")
        sys.exit(1)
    except socket.error as e:
        click.secho("{}: {}".format(type(e).__name__, e), fg='red')
        click.secho("Problem connecting to the Remote API socket. Check that the Remote API is enabled. "
                    "Wait one minute and try again. Also ensure the host setting is correct.")
        sys.exit(1)
    else:
        return netlab_client


def output(data, tablefmt='pipe', headers=(), **kwargs):
    """ Helper function to output data in tables. Accepts tabulate and click.secho options. """
    try:
        click.secho(tabulate(data.items(), tablefmt=tablefmt, headers=headers), **kwargs)
    except AttributeError:
        click.secho(tabulate(data, tablefmt=tablefmt, headers=headers), **kwargs)



import csv
import io

import click

from .utils import client, output
from ..constants import NL_OK
from ..errors.common import NetlabError
from ..serializer import serialize


@click.command()
@click.pass_context
@click.argument('method', nargs=1)
@click.argument('arguments', nargs=-1)
@click.option('--system', default='default', help="NETLAB+ system to communicate with.")
@click.option('--output', 'output_type', default='table', help="Change the output format. "
                                                               "[table, csv, tsv, json, var] ")
def api(ctx, method, arguments, system, output_type):
    """ Communicate with a remote NETLAB+ server API"""
    try:
        api_method = getattr(client(system), method)
    except AttributeError:
        api_method = None
        click.echo("{} is not a valid API method. Please check the documentation.".format(
            click.style(method, fg='red')))
        ctx.exit(1)

    api_arguments = {}
    try:
        for arg in arguments:
            key, value = arg.split('=')
            api_arguments[key.lower()] = value
    except ValueError:
        # user probably didn't pass arguments in the key=value format
        click.secho("There is an error with your method arguments.", fg='red')
        click.secho("Please ensure they are in the format of arg=value. \n\nExample:")
        click.secho("    {} user_account_get acc_id=100001".format(ctx.command_path), fg='cyan')
        ctx.exit(1)

    try:
        results = api_method(**api_arguments)
    except (TypeError, NetlabError) as err:
        results = None
        click.secho("{}: {}".format(type(err).__name__, err), fg='red')
        click.echo("Please check the documentation for help.")
        ctx.exit(1)

    if results == NL_OK:
        click.secho("Success!", fg='green')
    elif isinstance(results, int) or isinstance(results, str):
        click.echo(results)
    else:
        outputs = {
            'table': table_output,
            'csv': csv_output,
            'tsv': tsv_output,
            'json': json_output,
            'var': var_output
        }
        outputs[output_type](results)


def table_output(results):
    if isinstance(results, dict):
        results = [results]
    output(results, headers='keys')


def csv_output(results, delimiter=','):
    out = io.StringIO()

    if isinstance(results, dict):
        results = [results]

    fieldnames = list(results[0].keys())

    writer = csv.DictWriter(out, fieldnames=fieldnames, quoting=csv.QUOTE_MINIMAL, delimiter=delimiter)
    writer.writeheader()
    for result in results:
        writer.writerow(result)
    click.echo(out.getvalue())


def tsv_output(results):
    csv_output(results, delimiter='\t')


def json_output(results):
    click.echo(serialize(results, pretty=True))


def var_output(results):
    if isinstance(results, dict):
        results = [results]

    out = []
    for result in results:
        row = ""
        for key, value in result.items():
            row += "{}={} ".format(key, value)
        out.append(row)
    click.echo("\n".join(out))



import json
import logging
import os
from collections import namedtuple

from .errors.common import InvalidConfig

ENV_PREFIX = 'NETLAB_CONFIG_'
ENV_BOOLEANS = ['ssl']
CONFIG_FILENAME = os.path.join('.netlab', 'config.json')
DEFAULT_CONFIG_PATH = os.path.join(os.path.expanduser('~'), CONFIG_FILENAME)
log = logging.getLogger(__name__)

SystemConfig = namedtuple("SystemConfig", ['user', 'token', 'host', 'port', 'timeout', 'ssl'])


def get_system_config(system, config=None, config_path=None):
    defaults = {
        'host': 'localhost',
        'port': 9000,
        'timeout': 10,
        'ssl': True
    }

    if config:
        # config from Client init
        defaults.update(config)
    else:
        # file config
        defaults.update(load_config(system, config_path))

        # environment variables config
        env_vars = [var for var in os.environ.keys() if var.startswith(ENV_PREFIX)]
        if env_vars:
            env_config = {}
            for env in env_vars:
                key = env.replace(ENV_PREFIX, '').lower()
                if key in ENV_BOOLEANS:
                    # convert env value to python boolean
                    env_config[key] = os.environ.get(env).lower() in ("true", "yes", "t", "1")
                else:
                    env_config[key] = os.environ.get(env)
            defaults.update(env_config)

    if 'user' not in defaults:
        raise InvalidConfig('Config missing required "user".')
    if 'token' not in defaults:
        raise InvalidConfig('Config missing required "token".')
    return SystemConfig(**defaults)


def load_config(system=None, config_path=None):
    """
    Loads authentication data from a Netlab configuration file in the given
    root directory or if config_path is passed use given path.
    If no system is specified, return entire config dict.
    Lookup priority:
        explicit config_path parameter > NETLAB_CONFIG environment variable > ~/.netlab/config.json
    """
    config_file = find_config_file(config_path)

    if not config_file:
        return {}

    try:
        with open(config_file) as f:
            data = json.load(f)
            if system:
                return data.get(system, {})
            else:
                return data
    except (IOError, KeyError, ValueError) as e:
        log.debug(e)
        raise


def add_config(system, config, config_path=None):
    """
    Adds system and config data from a Netlab to the configuration file
    """
    current_config = load_config(config_path=config_path)
    current_config[system] = config
    config_file = find_config_file(config_path)
    if not config_file:
        config_file = DEFAULT_CONFIG_PATH

    try:
        os.makedirs(os.path.dirname(config_file), exist_ok=True)
        with open(config_file, 'w') as f:
            json.dump(current_config, f, sort_keys=True, indent=4)
    except (IOError, KeyError, ValueError) as e:
        log.debug(e)
        raise


def remove_config(system, config_path=None):
    """
    Removes system and config data from the configuration file
    """
    current_config = load_config(config_path=config_path)
    if not system:
        current_config = {}
    else:
        current_config.pop(system, None)

    config_file = find_config_file(config_path)

    try:
        with open(config_file, 'w') as f:
            json.dump(current_config, f, sort_keys=True, indent=4)
    except (IOError, KeyError, ValueError) as e:
        log.debug(e)
        raise


def find_config_file(config_path=None):
    environment_path = os.path.join(
        os.environ.get('NETLAB_CONFIG'),
        os.path.basename(CONFIG_FILENAME)
    ) if os.environ.get('NETLAB_CONFIG') else None

    paths = filter(None, [
        config_path,
        environment_path,
        DEFAULT_CONFIG_PATH
    ])

    for path in paths:
        if os.path.exists(path):
            return path

    return None



import logging
import time

from .api import ClassApiMixin, PodApiMixin, ReservationApiMixin, SystemApiMixin, UserApiMixin, VmApiMixin
from .auth import get_system_config
from .connection import Connection, SSLConnection
from .errors import error_decode, ResponseFormatError
from .serializer import serialize, deserialize

log = logging.getLogger(__name__)


class Client(ClassApiMixin, PodApiMixin, ReservationApiMixin, SystemApiMixin, UserApiMixin, VmApiMixin):
    def __init__(self, system='default', config=None, config_path=None):
        """
        A client for communicating with a NETLAB+ server.

        :param str system: The name to identify a NETLAB+ system. Defaults to **'default'**.
        :param dict config: Configuration options. See below for all possible options for the :attr:`config`.
        :param str config_path: File path location of a JSON config file to be used instead of the default location.

        :returns: A client instance with which you can use to interact with a remote NETLAB+ system.
        """
        self.id = 0
        self.system = system
        self.config = get_system_config(system, config, config_path)

        if self.config.ssl:
            connection_class = SSLConnection
        else:
            connection_class = Connection

        self.connection = connection_class(self.config.user, self.config.token, self.config.host, self.config.port,
                                           socket_timeout=self.config.timeout)
        self.connection.connect()

    def __repr__(self):
        return "NetlabClient<system={system},user={user},host={host}>".format(
            system=self.system, **self.config._asdict()
        )

    def call(self, method, task_wait=True, task_poll=3, **kwargs):
        """
        Calls an API method.

        .. warning::
            This is a low-level API method and should not be used by most users.

        **Required:**

        :param str method: the method name

        **Optional:**

        :param bool task_wait: wait on task True/False
        :param int task_poll: number of seconds between task polling (default 3)

        **Keyword Arguments:**

        All other named arguments are passed as parameters to the
        API backend and are specific to the method type.  See the API
        documentation for method parameters.

        **Tasks:**

        Methods ending in ".task" are long running tasks.  Tasks may
        be invoked syncronously or asyncronously by setting task_wait.

        * If *task_wait == True*, call() will continuously poll the task,
          wait for completion or error, and return task result on success.
          This is the default if task_wait is not specified.
        * If *task_wait == False*, call() will return the task_id of the
          created task but will not wait.  This may be helpful if you
          want to do other work while waiting, or run tasks in parallel.
          Call task_check() to get the current task state, task progress,
          task result, or task error.
          Call task_wait() at a later time to wait on a task.

        :returns: A value or an object (check the API documentation).

        :raises NetlabError: if the NETLAB+ server returns an error response.
        """
        if method.endswith('.task'):
            if type(task_wait) is not bool:
                raise ResponseFormatError('argument "task_wait" must be bool')
            if type(task_poll) is not int or task_poll < 0:
                raise ResponseFormatError('task_poll: positive int required')
        else:
            task_wait = False

        if self.id < 2 ^ 62:
            self.id += 1
        else:
            self.id = 1
        rpc = {
            'id': self.id,
            'jsonrpc': '2.0',
            'method': method,
            'params': kwargs
        }
        request = serialize(rpc)
        log.debug("send %s", request)
        self.connection.send(request)

        raw_response = self.connection.read_response()
        log.debug("recv %s", raw_response)
        response = deserialize(raw_response)

        if type(response) is not dict:
            raise ResponseFormatError('Received invalid response, not dict')
        if response['id'] != self.id:
            raise ResponseFormatError('Reply id did not match request')
        if 'result' in response:
            if task_wait:
                task_id = response['result']
                result = self.task_wait(task_id, task_poll)
                return result
            else:
                return response['result']
        if 'error' in response:
            raise error_decode(response['error'])
        raise ResponseFormatError('Response did not contain result or error')

    def task_wait(self, task_id, task_poll=3):
        if task_poll <= 0:
            raise ResponseFormatError('task_poll must be int > 0')
        while True:
            status = self.task_check(task_id)
            if status['is_complete']:
                if status['error'] is not None:
                    raise error_decode(status['error'])
                return status['result']
            time.sleep(task_poll)

    def task_check(self, task_id):
        return self.call('task.check', task_id=task_id)



from .common import NetlabError

CODES = []


def _add(cls):
    CODES.append(cls)
    return cls


@_add
class VirtualMachineIsTemplateError(NetlabError):
    """Virtual machine role is a template."""
    str_error = "E_VM_IS_TEMPLATE"


@_add
class VirtualMachineIsPersistentError(NetlabError):
    """Virtual machine role is persistent."""
    str_error = "E_VM_IS_PERSISTENT"


@_add
class VirtualMachineIsRuntimeError(NetlabError):
    """Virtual machine role is runtime."""
    str_error = "E_VM_IS_RUNTIME"


@_add
class VirtualMachineSnapshotRequiredError(NetlabError):
    """snapshot_id or snapshot_name is required."""
    str_error = "E_VM_SNAPSHOT_REQUIRED"


@_add
class VirtualMachineToolsUnavailableError(NetlabError):
    """VMWare tools are unavailable."""
    str_error = "E_VM_TOOLS_UNAVAILABLE"


@_add
class VirtualMachineShutdownTimeoutError(NetlabError):
    """The virtual machine did not shutdown within wait_seconds seconds."""
    str_error = "E_VM_SHUTDOWN_TIMEOUT"


@_add
class VirtualMachineInvalidHostError(NetlabError):
    """Could not determine runtime host from virtual machine."""
    str_error = "E_VM_INVALID_HOST"


@_add
class VirtualMachineHostNotFoundError(NetlabError):
    """vh_id not found."""
    str_error = "E_VM_HOST_NOT_FOUND"


@_add
class InternalError(NetlabError):
    """VM pc_id not the same as inventory pc_id."""
    str_error = "E_INTERNAL_ERROR"


@_add
class VirtualMachineWrongHostError(NetlabError):
    """vm_id is powered up on incorrect host."""
    str_error = "E_VM_WRONG_HOST"


@_add
class VirtualMachineInvalidPowerStateError(NetlabError):
    """If ok_if_on is not set, then VM was already powered on."""
    str_error = "E_VM_INVALID_POWER_STATE"


@_add
class VirtualMachineDatacenterMismatchError(NetlabError):
    """vm_id and vh_id not in the same datacenter."""
    str_error = "E_VM_DATACENTER_MISMATCH"


@_add
class InvalidParametersError(NetlabError):
    """"""
    str_error = "E_INVALID_PARAMETERS"


@_add
class PCNotFoundError(NetlabError):
    """pc_id is not found."""
    str_error = "E_PC_NOT_FOUND"


@_add
class VirtualMachineNotFoundError(NetlabError):
    """vm_id could not be found."""
    str_error = "E_VM_NOT_FOUND"


@_add
class VirtualMachineDatacenterNotFoundError(NetlabError):
    """vdc_id not found."""
    str_error = "E_VM_DATACENTER_NOT_FOUND"


@_add
class VirtualMachineDatacenterNotFoundUnique(NetlabError):
    """vdc_id not unique."""
    str_error = "E_VM_DATACENTER_NOT_UNIQUE"


@_add
class VirtualMachineInvalidStateError(NetlabError):
    """Virtual machine is not currently connected."""
    str_error = "E_VM_INVALID_STATE"


@_add
class VirtualMachineNotUniqueError(NetlabError):
    """Duplicate vm_uuid in datacenter vdc_id."""
    str_error = "E_VM_NOT_UNIQUE"


@_add
class VirtualMachineParentNotFoundError(NetlabError):
    """vm_parent_id is not found."""
    str_error = "E_VM_PARENT_NOT_FOUND"


@_add
class VirtualMachineHostInvalidNameError(NetlabError):
    """vh_name is invalid."""
    str_error = "E_VM_HOST_INVALID_NAME"


@_add
class VirtualMachineHostNotUniqueError(NetlabError):
    """vh_name is not unique."""
    str_error = "E_VM_HOST_NOT_UNIQUE"


@_add
class DataBaseUpdateFailedError(NetlabError):
    """Database error."""
    str_error = "E_DB_UPDATE_FAILED"


@_add
class ValueRequiredError(NetlabError):
    """vdc_name is required."""
    str_error = "E_VALUE_REQUIRED"


@_add
class ValueIsBlankError(NetlabError):
    """vdc_name can not be blank."""
    str_error = "E_VALUE_IS_BLANK"


@_add
class ValueNotWebsafeError(NetlabError):
    """vdc_name is not web safe."""
    str_error = "E_VALUE_NOT_WEBSAFE"


@_add
class VDCMONStopError(NetlabError):
    """Could not stop vdcmon for vdc_id."""
    str_error = "E_VDCMON_STOP"


@_add
class VirtualMachineDatacenterInvalidHostnameError(NetlabError):
    """vdc_agent_hostname is invalid."""
    str_error = "E_VM_DATACENTER_INVALID_HOSTNAME"


@_add
class VirtualMachineDatacenterInvalidTypeError(NetlabError):
    """vdt_type is invalid."""
    str_error = "E_VM_DATACENTER_INVALID_TYPE"


@_add
class VirtualMachineNotAVMI(NetlabError):
    """Virtual machine is a template."""
    str_error = "E_VM_NOT_AVMI"


@_add
class VirtualMachineSnapshotCurrentNotFoundError(NetlabError):
    """Could not revert VM image to current snapshot, a current snapshot does not exist."""
    str_error = "E_VM_SNAPSHOT_CURRENT_NOT_FOUND"


@_add
class VirtualMachineHostPortGroupNotFoundError(NetlabError):
    """Could not map portgroup.id to portgroup name. Portgroup does not exist on host."""
    str_error = "E_VM_HOST_PORTGROUP_NOT_FOUND"


@_add
class VirtualMachineInvalidDeviceError(NetlabError):
    """Network adapter vnic does not exist on this vm."""
    str_error = "E_VM_INVALID_DEVICE"


@_add
class VirtualMachineHostGroupNotMemberError(NetlabError):
    """clone_vh_id is not a part of clone_vhg_id."""
    str_error = "E_VM_HOST_GROUP_NOT_MEMBER"


@_add
class VirtualMachineHostGroupEmptyError(NetlabError):
    """clone_vhg_id is empty."""
    str_error = "E_VM_HOST_GROUP_EMPTY"


@_add
class DatacenterMismatchError(NetlabError):
    """Host clone_vh_id is not in parent datacenter vdc_id."""
    str_error = "E_DATACENTER_MISMATCH"



from .common import NetlabError

CODES = []


def _add(cls):
    CODES.append(cls)
    return cls


@_add
class ContentNotFoundError(NetlabError):
    """con_id is not found."""
    str_error = "E_CONTENT_NOT_FOUND"


@_add
class ExerciseNotFoundError(NetlabError):
    """ex_id is not found."""
    str_error = "E_EXERCISE_NOT_FOUND"



class ClassApiMixin:
    class_default_props = ['cls_id', 'cls_name', 'com_id']
    class_props = ['cls_change_ex', 'cls_def_pw_account', 'cls_def_pw_console', 'cls_def_pw_enable',
                   'cls_email_logs', 'cls_end_date', 'cls_id', 'cls_lab_limit', 'cls_max_slots_per_res',
                   'cls_min_hours_btw_res', 'cls_name', 'cls_no_delete', 'cls_retain_ilt', 'cls_retain_period',
                   'cls_retain_st', 'cls_self_sched', 'cls_start_date', 'cls_team_sched', 'cls_uuid', 'com_id',
                   'enrollment', 'leads']

    roster_default_props = ['acc_id', 'acc_user_id', 'acc_email', 'lead']
    roster_props = ['acc_display_name', 'acc_email', 'acc_full_name', 'acc_id', 'acc_last_login', 'acc_logins',
                    'acc_sort_name', 'acc_time_last_access', 'acc_type', 'acc_user_id', 'cls_id', 'lead', 'ros_team']

    def class_list(self, *, com_id=None, member=False, properties="default", **kwargs):
        """
        This method allows you to retrieve a list of classes and their properties.

        :param str com_id: Community ID.
        :param bool member: If true, get list of classes the caller is in.
        :param properties: List of properties to retrieve. See Returns section for list of available properties.
        :type properties: str or list

        :return: On success, returns an object with the requested properties:
        :rtype: dict

        **Properties**

        cls_change_ex
            Allow student to change exercise.
        cls_def_pw_account
            Class default account password.
        cls_def_pw_console
            Class default console password.
        cls_def_pw_enable
            Class default enable password.
        cls_email_logs
            Email reservation logs for real equipment pods. (N=No, P=Email primary instructor)
        cls_end_date
            Class end date.
        cls_id
            Unique class identifier.
        cls_lab_limit
            Lab limit settings. (E=Enforce class lab limit, I=Instructor override)
        cls_max_slots_per_res
            Maximum number of slots per reservation.
        cls_min_hours_btw_res
            Minimum number of hours between reservations.
        cls_name
            Unique class name per community.
        cls_no_delete
            Prevent from deletion.
        cls_retain_ilt
            Retain instructor led logs.
        cls_retain_period
            Number of days to retain class logs.
        cls_retain_st
            Retain student logs.
        cls_self_sched
            Allow students to self schedule.
        cls_start_date
            Class start date.
        cls_team_sched
            Allow teams to schedule.
        cls_uuid
            Class UUID.
        com_id
            Unique community identifier.
        enrollment
            Count of those enrolled.
        leads
            List of leads and properties sorted by acc_sort_name.
        """
        method = "class.list"
        if properties == 'all':
            properties = self.class_props
        if properties == 'default':
            properties = self.class_default_props
        return self.call(method, com_id=com_id, member=member, properties=properties, **kwargs)

    def class_add(self, *, cls_name, com_id, cls_change_ex=True, cls_def_pw_account=None, cls_def_pw_console=None,
                  cls_def_pw_enable=None, cls_email_logs="N", cls_end_date=None, cls_lab_limit="E",
                  cls_max_slots_per_res=None, cls_min_hours_btw_res=None, cls_no_delete=False, cls_retain_ilt=False,
                  cls_retain_period=None, cls_retain_st=True, cls_self_sched=True, cls_start_date=None,
                  cls_team_sched=False, **kwargs):
        """
        This method allows you to add a class to the NETLAB+ system.

        :param str cls_name: Class name.
        :param int com_id: Community ID.

        :param bool cls_change_ex: Allow student to change exercise.
        :param str cls_def_pw_account: Class default account password.
        :param str cls_def_pw_console: Class default console password.
        :param str cls_def_pw_enable: Class default enable password.
        :param str cls_email_logs: Email reservation logs for real equipment pods:
            'N' - Don't email logs 'P' - Email to primary instructor
        :param datetime.date cls_end_date: Class end date.
        :param str cls_lab_limit: Class lab time limit:
            'E' - Enforce time limit or 'I' - Ignore time limit.
        :param int cls_max_slots_per_res: Maximum number of slots per reservation.
        :param int cls_min_hours_btw_res: Minimum hours between reservations.
        :param bool cls_no_delete: Prevent from deletion.
        :param bool cls_retain_ilt: Retain instructor led logs.
        :param int cls_retain_period: The number of days to retain logs for.
        :param bool cls_retain_st: Retain student logs.
        :param bool cls_self_sched: Allow students to self schedule.
        :param datetime.date cls_start_date: Class start date.
        :param bool cls_team_sched: Allow teams to schedule.

        :return: On success, returns the Class ID of the newly created class.
        :rtype cls_id: int
        """

        method = "class.add"
        params = {'cls_name': cls_name, 'com_id': com_id, 'cls_change_ex': cls_change_ex,
                  'cls_email_logs': cls_email_logs.upper(), 'cls_lab_limit': cls_lab_limit.upper(),
                  'cls_no_delete': cls_no_delete, 'cls_retain_ilt': cls_retain_ilt, 'cls_retain_st': cls_retain_st,
                  'cls_self_sched': cls_self_sched, 'cls_team_sched': cls_team_sched}

        if cls_def_pw_account:
            params['cls_def_pw_account'] = cls_def_pw_account
        if cls_def_pw_console:
            params['cls_def_pw_console'] = cls_def_pw_console
        if cls_def_pw_enable:
            params['cls_def_pw_enable'] = cls_def_pw_enable

        if cls_start_date:
            params['cls_start_date'] = cls_start_date
        if cls_end_date:
            params['cls_end_date'] = cls_end_date

        if cls_max_slots_per_res:
            params['cls_max_slots_per_res'] = cls_max_slots_per_res
        if cls_min_hours_btw_res:
            params['cls_min_hours_btw_res'] = cls_min_hours_btw_res

        if cls_retain_period:
            params['cls_retain_period'] = cls_retain_period

        params.update(kwargs)
        return int(self.call(method, **params))

    def class_get(self, *, cls_id, properties="default", **kwargs):
        """
        This method allows you to retrieve a single class and properties.

        :param str cls_id: Class ID.
        :type cls_id: str
        :param properties: List of properties to retrieve. Defaults to "all"
        :type properties: str or list
        :return: On success, returns an object with the requested properties:
        :rtype: dict

        **Properties**

        cls_change_ex
            Allow student to change exercise.
        cls_def_pw_account
            Class default account password.
        cls_def_pw_console
            Class default console password.
        cls_def_pw_enable
            Class default enable password.
        cls_email_logs
            Email reservation logs for real equipment pods. (N=No, P=Email primary instructor)
        cls_end_date
            Class end date.
        cls_id
            Unique class identifier.
        cls_lab_limit
            Lab limit settings. (E=Enforce class lab limit, I=Instructor override)
        cls_max_slots_per_res
            Maximum number of slots per reservation.
        cls_min_hours_btw_res
            Minimum number of hours between reservations.
        cls_name
            Unique class name per community.
        cls_no_delete
            Prevent from deletion.
        cls_retain_ilt
            Retain instructor led logs.
        cls_retain_period
            Number of days to retain class logs.
        cls_retain_st
            Retain student logs.
        cls_self_sched
            Allow students to self schedule.
        cls_start_date
            Class start date.
        cls_team_sched
            Allow teams to schedule.
        cls_uuid
            Class UUID.
        com_id
            Unique community identifier.
        enrollment
            Count of those enrolled.
        leads
            List of leads and properties sorted by acc_sort_name.
        """

        method = "class.get"
        if properties == 'all':
            properties = self.class_props
        if properties == 'default':
            properties = self.class_default_props
        return self.call(method, cls_id=cls_id, properties=properties, **kwargs)

    def class_update(self, *, cls_id, cls_change_ex=None, cls_def_pw_account=None, cls_def_pw_console=None,
                     cls_def_pw_enable=None, cls_email_logs=None, cls_end_date=None, cls_lab_limit=None,
                     cls_max_slots_per_res=None, cls_min_hours_btw_res=None, cls_name=None, cls_no_delete=None,
                     cls_retain_ilt=None, cls_retain_period=None, cls_retain_st=None, cls_self_sched=None,
                     cls_start_date=None, cls_team_sched=None, **kwargs):
        """
        This method allows you to update a class and its properties.

        :param str cls_id: Unique class identifier.

        :param bool cls_change_ex: Allow student to change exercise.
        :param str cls_def_pw_account: Class default account password.
        :param str cls_def_pw_console: Class default console password.
        :param str cls_def_pw_enable: Class default enable password.
        :param str cls_email_logs: Email reservation logs for real equipment pods:
                               'N' - Don't email logs
                               'P' - Email to primary instructor
        :param datetime.date cls_end_date: Class end date.
        :param str cls_lab_limit: Class lab time limit: 'E' - Enforce time limit or 'I' - Ignore time limit.
        :param int cls_max_slots_per_res: Maximum number of slots per reservation.
        :param int cls_min_hours_btw_res: Minimum hours between reservations.
        :param str cls_name: Class name.
        :param bool cls_no_delete: Prevent from deletion.
        :param bool cls_retain_ilt: Retain instructor led logs.
        :param int cls_retain_period: The number of days to retain logs for.
        :param bool cls_retain_st: Retain student logs.
        :param bool cls_self_sched: Allow students to self schedule.
        :param datetime.date cls_start_date: Class start date.
        :param bool cls_team_sched: Allow teams to schedule.

        :return: OK or Error
        """

        method = "class.update"
        params = {'cls_id': cls_id}

        if cls_name:
            params['cls_name'] = cls_name
        if cls_start_date:
            params['cls_start_date'] = cls_start_date
        if cls_end_date:
            params['cls_end_date'] = cls_end_date
        if cls_def_pw_account:
            params['cls_def_pw_account'] = cls_def_pw_account
        if cls_def_pw_console:
            params['cls_def_pw_console'] = cls_def_pw_console
        if cls_def_pw_enable:
            params['cls_def_pw_enable'] = cls_def_pw_enable
        if cls_email_logs:
            params['cls_email_logs'] = cls_email_logs
        if cls_lab_limit:
            params['cls_lab_limit'] = cls_lab_limit
        if cls_no_delete is not None:
            params['cls_no_delete'] = cls_no_delete
        if cls_retain_ilt is not None:
            params['cls_retain_ilt'] = cls_retain_ilt
        if cls_retain_st is not None:
            params['cls_retain_st'] = cls_retain_st
        if cls_retain_period:
            params['cls_retain_period'] = cls_retain_period
        if cls_self_sched is not None:
            params['cls_self_sched'] = cls_self_sched
        if cls_team_sched is not None:
            params['cls_team_sched'] = cls_team_sched
        if cls_change_ex is not None:
            params['cls_change_ex'] = cls_change_ex
        if cls_min_hours_btw_res:
            params['cls_min_hours_btw_res'] = cls_min_hours_btw_res
        if cls_max_slots_per_res:
            params['cls_max_slots_per_res'] = cls_max_slots_per_res

        params.update(kwargs)
        return self.call(method, **params)

    def class_remove(self, *, cls_id, delete_students=False, **kwargs):
        """
        This method allows you to remove a class from the NETLAB+ system.

        :param str cls_id: Unique class identifier.
        :param bool delete_students: Delete student accounts as well.

        :return: OK or Error
        """

        method = "class.remove"
        return self.call(method, cls_id=cls_id, delete_students=delete_students, **kwargs)

    def class_roster_add(self, *, acc_id, cls_id, lead=False, ros_team=None, **kwargs):
        """
        This method allows you to add a member to a class roster.

        :param str acc_id: Unique account identifier.
        :param str cls_id: Unique class identifier.
        :param bool lead: Add as lead.
        :param str ros_team: Assign user to a team. Single character, A-Z.

        :return: OK or Error
        """

        method = "class.roster.add"
        params = {'acc_id': acc_id, 'cls_id': cls_id, 'lead': lead}

        if ros_team:
            params['ros_team'] = ros_team

        params.update(kwargs)
        return self.call(method, **params)

    def class_roster_list(self, *, cls_id, leads=False, properties="default", **kwargs):
        """
        This method lists users in the class roster.

        :param str cls_id: Unique class identifier.
        :param bool leads: Retrieve **only** leads.
        :param properties: List of properties to retrieve.
        :type properties: str, list

        :return: On success, returns an object with the requested properties:

        **Properties**

        acc_display_name
            Account name people will see this name in discussions, messages and comments.
        acc_email
            Account email.
        acc_full_name
            Account full name.
        acc_last_login
            Timestamp of last account login.
        acc_logins
            Number of times a user has logged in.
        acc_sort_name
            Account name as it appears in sorted lists.
        acc_time_last_access
            Timestamp of last access.
        acc_type
            Account type: 'S': Student, 'I': Instructor
        acc_user_id
            Account user identifier.
        """

        method = "class.roster.list"
        if properties == 'all':
            properties = self.roster_props
        if properties == 'default':
            properties = self.roster_default_props
        return self.call(method, cls_id=cls_id, leads=leads, properties=properties, **kwargs)

    def class_roster_remove(self, *, acc_id, cls_id, **kwargs):
        """
        This method removes a user from the class roster.

        :param str acc_id: Account identifier of user to be removed.
        :param str cls_id: Unique class identifier.

        :return: OK or Error
        """

        method = "class.roster.remove"
        return self.call(method, acc_id=acc_id, cls_id=cls_id, **kwargs)



class UserApiMixin:
    user_account_default_props = ['acc_id', 'acc_user_id', 'acc_email', 'acc_type']
    user_account_list_props = ['acc_can_login', 'acc_display_name', 'acc_email', 'acc_fs_id', 'acc_full_name', 'acc_id',
                               'acc_last_ip', 'acc_last_login', 'acc_last_ua', 'acc_login_queue',
                               'acc_logins', 'acc_privs', 'acc_pw_change', 'acc_session_time_started', 'acc_sort_name',
                               'acc_sys', 'acc_time_created', 'acc_time_last_access', 'acc_type', 'acc_user_id',
                               'acc_uuid', 'com_id', 'tz_id']
    user_account_get_props = user_account_list_props + ['com_full_name', 'tz_iana', 'tz_name']

    user_account_privs = ['COMWIDE', 'SYSWIDE', 'POD_DESIGNER', 'LAB_DESIGNER']

    community_default_props = ['com_id', 'com_full_name']
    community_props = ['com_alt_banner', 'com_alt_banner_height', 'com_alt_banner_width', 'com_enabled',
                       'com_full_name', 'com_id', 'com_max_slots_per_res', 'com_min_hours_btw_res',
                       'com_mynetlab_news', 'com_mynetlab_welcome']

    def user_account_list(self, *, com_id=None, acc_type=None, properties="default", **kwargs):
        """
        Query a list of the user accounts in the NETLAB system. The list method
        is only restricted by community id and account types. The properties
        only identify those properties or fields that will be returned via
        the underlying query.

        :param int com_id: Community ID
        :param str acc_type: User account type **(S, I, None)**
        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.
        :type properties: str or list

        :returns: A list of dictionaries (a set of records) where the dictionaries contain requested properties
            per user.

        **Properties**

        acc_can_login
            User allowed to login.
        acc_display_name
            Account name people will see this name in discussions, messages and comments.
        acc_email
            Account email.
        acc_fs_id
            Filesystem identifier of the account root folder.
        acc_full_name
            Account full name.
        acc_id
            Account identifier.
        acc_last_ip
            Last IP address of user.
        acc_last_login
            Datetime of last account login.
        acc_last_ua
            User agent string of the browser when logging in from the UI.
        acc_logins
            Number of account logins.
        acc_privs
            Supplemental account privileges for instructors only.
            **('COMWIDE', 'SYSWIDE', 'POD_DESIGNER', 'LAB_DESIGNER')**
        acc_pw_change
            Force user to change password.
        acc_session_time_started
            Datetime of session start time.
        acc_sort_name
            Account name as it appears in sorted lists.
        acc_sys
            Mark as system account.
        acc_time_created
            Datetime account was created.
        acc_time_last_access
            Datetime of last access.
        acc_type
            Account type. 'S' (student), 'I' (instructor), 'Z' (administrator)
        acc_user_id
            Unique system-wide login identifier.
        acc_uuid
            Globally unique account ID.
        com_id
            Unique community identifier.
        tz_id
            Time zone identifier.
        """

        method = 'user.account.list'

        if acc_type:
            acc_type = acc_type.upper()
        if properties == 'all':
            properties = self.user_account_list_props
        if properties == 'default':
            properties = self.user_account_default_props
        return self.call(method, com_id=com_id, acc_type=acc_type, properties=properties, **kwargs)

    def user_account_filter(self, *, com_id=None, filter_params, **kwargs):
        """
        Gets all users from server and then performs filtering locally. Any property from
        :meth:`~netlab.api.UserApiMixin.user_account_list` can be filtered upon. Multiple
        ``filter_params`` will be combined using AND. Returns zero or more user accounts that match filter query.

        :param int com_id: Only return users from this Community ID, defaults to all.
        :param dict filter_params: Dict where each key is a property and the value to filter by.

        :returns: A list of user accounts that match ``filter_params``
        """

        # only request the properties that are needed to filter by and always get acc_id
        properties = list(filter_params) + ['acc_id']

        # TODO Replace with server side filtering when available. client side could be very slow
        def filter_by_params(account):
            filter_checks = []
            for param, value in filter_params.items():
                # TODO: add filter_param__options like case insensitive, date range, etc
                filter_checks.append(account[param] == value)
            return all(filter_checks)

        users = self.call('user.account.list', com_id=com_id, properties=properties, **kwargs)
        results = list(filter(filter_by_params, users))
        return results

    def user_account_get(self, *, acc_id, properties="default", **kwargs):
        """
        Retrieves a user account on the NETLAB system.

        :param int acc_id: User Account ID.
        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.
        :type properties: str or list

        :returns: Account data for the requested acc_id with the requested properties.

        **Properties**

        acc_can_login
            User allowed to login.
        acc_display_name
            Account name people will see this name in discussions, messages and comments.
        acc_email
            Account email.
        acc_fs_id
            Filesystem identifier of the account root folder.
        acc_full_name
            Account full name.
        acc_id
            Account identifier.
        acc_last_ip
            Last IP address of user.
        acc_last_login
            Datetime of last account login.
        acc_last_ua
            User agent string of the browser when logging in from the UI.
        acc_logins
            Number of account logins.
        acc_privs
            Supplemental account privileges for instructors only.
            **('COMWIDE', 'SYSWIDE', 'POD_DESIGNER', 'LAB_DESIGNER')**
        acc_pw_change
            Force user to change password.
        acc_session_time_started
            Datetime of session start time.
        acc_sort_name
            Account name as it appears in sorted lists.
        acc_sys
            Mark as system account.
        acc_time_created
            Datetime account was created.
        acc_time_last_access
            Datetime of last access.
        acc_type
            Account type. 'S' (student), 'I' (instructor), 'Z' (administrator)
        acc_user_id
            Unique system-wide login identifier.
        acc_uuid
            Globally unique account ID.
        com_full_name
            Community name.
        com_id
            Unique community identifier.
        tz_iana
            Timezone IANA database identifier.
        tz_id
            Time zone identifier.
        tz_name
            Time zone name.
        """
        method = 'user.account.get'

        if properties == 'all':
            properties = self.user_account_get_props
        if properties == 'default':
            properties = self.user_account_default_props

        return self.call(method, acc_id=acc_id, properties=properties, **kwargs)

    def user_account_add(self, *, com_id, acc_user_id, acc_password, acc_full_name, acc_type="S", acc_privs=None,
                         acc_display_name=None, acc_sort_name=None, acc_email=None, acc_can_login=True,
                         acc_pw_change=True, cls_id=None, tz_id=11, **kwargs):
        """
        Add a user account on the NETLAB system.

        :param int com_id: Community ID.
        :param str acc_user_id: User Account ID.
        :param str acc_password: Unencrypted account password.
        :param str acc_full_name: Account full name.

        :param str acc_type: Account type Student = 'S' or Instructor = 'I'. Defaults to 'S'.
        :param bool acc_can_login: User can login. Defaults to True.
        :param bool acc_pw_change: Upon first login, force user to change password. Defaults to True.
        :param str acc_privs: Supplemental account privileges for instructors in list or tuple format.
                              **('COMWIDE', 'SYSWIDE', 'POD_DESIGNER', 'LAB_DESIGNER')**
        :param str acc_display_name: Account name, people will see this name in discussions, messages, and comments.
        :param str acc_sort_name: Account name as it appears in sorted lists.
        :param str acc_email: Account email address.
        :param int tz_id: Timezone identifier. Defaults to 11.
        :param int cls_id: Class ID. (If you want to add user directly to a class during account creation.)

        :returns: On success, returns acc_id.
        :rtype: int
        """

        method = "user.account.add"
        params = {
            'com_id': com_id,
            'acc_user_id': acc_user_id,
            'acc_password': acc_password,
            'acc_full_name': acc_full_name,
            'acc_type': acc_type.upper(),
            'acc_can_login': acc_can_login,
            'acc_pw_change': acc_pw_change,
            'tz_id': tz_id,
        }

        if acc_privs:
            if isinstance(acc_privs, (list, tuple)):
                acc_privs = ",".join(acc_privs)
            params['acc_privs'] = acc_privs.upper()
        if acc_display_name:
            params['acc_display_name'] = acc_display_name
        if acc_sort_name:
            params['acc_sort_name'] = acc_sort_name
        if acc_email:
            params['acc_email'] = acc_email
        if cls_id:
            params['cls_id'] = cls_id

        params.update(kwargs)
        return int(self.call(method, **params))

    def user_account_password_set(self, *, acc_id, new_password, force_reset=False, **kwargs):
        """
        Sets the user account password.

        :param int acc_id: Account ID.
        :param str new_password: New account password.
        :param bool force_reset: Force user to reset password. Defaults to False.

        :returns: Acknowledgment of "OK" or "Error"
        """

        method = "user.account.password.set"
        return self.call(method, acc_id=acc_id, new_password=new_password, force_reset=force_reset, **kwargs)

    def user_account_remove(self, *, acc_id, **kwargs):
        """
        Removes the specified user account.

        :param int acc_id: User Account ID

        :returns: Ok or Error
        """

        method = "user.account.remove"
        return self.call(method, acc_id=acc_id, **kwargs)

    def user_account_update(self, *, acc_id, acc_user_id=None, acc_type=None, acc_can_login=None,
                            acc_pw_change=None, acc_password=None, acc_privs=None, acc_full_name=None,
                            acc_display_name=None, acc_sort_name=None, acc_email=None,
                            tz_id=None, **kwargs):
        """
        Update the user account.

        :param int acc_id: Account ID

        :param str acc_user_id: Unique system-wide login identifier.
        :param str acc_type: Account type. (Student='S', Instructor='I')
        :param bool acc_can_login: User allowed to login.
        :param bool acc_pw_change: Force user to change password.
        :param str acc_password: Account password
        :param list acc_privs: Supplemental account privileges for instructors in list or tuple format.
                               **('COMWIDE', 'SYSWIDE', 'POD_DESIGNER', 'LAB_DESIGNER')**
        :param str acc_full_name: Account full name.
        :param str acc_display_name: Account name people will see this name in discussions, messages and comments.
        :param str acc_sort_name: Account name as it appears in sorted lists.
        :param str acc_email: Account e-mail address.
        :param int tz_id: Timezone identifier.

        :returns: On success, returns 'OK'
        """
        method = "user.account.update"
        params = {'acc_id': acc_id}

        if acc_user_id:
            params['acc_user_id'] = acc_user_id
        if acc_type:
            params['acc_type'] = acc_type
        if acc_can_login is not None:
            params['acc_can_login'] = acc_can_login
        if acc_pw_change is not None:
            params['acc_pw_change'] = acc_pw_change
        if acc_password:
            params['acc_password'] = acc_password
        if acc_privs:
            if isinstance(acc_privs, (list, tuple)):
                acc_privs = ",".join(acc_privs)
            params['acc_privs'] = acc_privs.upper()
        if acc_full_name:
            params['acc_full_name'] = acc_full_name
        if acc_display_name:
            params['acc_display_name'] = acc_display_name
        if acc_sort_name:
            params['acc_sort_name'] = acc_sort_name
        if acc_email:
            params['acc_email'] = acc_email
        if tz_id:
            params['tz_id'] = tz_id

        params.update(kwargs)
        return self.call(method, **params)

    def user_community_add(self, *, com_full_name, com_id, com_enabled=True,
                           com_max_slots_per_res=None, com_min_hours_btw_res=None,
                           com_mynetlab_news=None, com_mynetlab_welcome=None):
        """
        This method allows you to add a community to the NETLAB+ system.

        :param str com_full_name: Community full name.
        :param int com_id: Community identifier.
        :param bool com_enabled: Is community enabled.
        :param int com_max_slots_per_res: Max amount of slots per reservation.
        :param int com_min_hours_btw_res: Minimum hours between reservations.
        :param str com_mynetlab_news: Community MyNETLAB+ news entry.
        :param str com_mynetlab_welcome: Community MyNETLAB+ welcome message.

        :returns: Community ID of the newly created community.
        """

        assert com_id < 100, "Community id out of range"  # TODO document and confirm
        params = {
            'com_full_name': com_full_name,
            'com_id': com_id,
            'com_enabled': com_enabled,
        }

        if com_max_slots_per_res is not None:
            params['com_max_slots_per_res'] = com_max_slots_per_res
        if com_min_hours_btw_res is not None:
            params['com_min_hours_btw_res'] = com_min_hours_btw_res
        if com_mynetlab_news is not None:
            params['com_mynetlab_news'] = com_mynetlab_news
        if com_mynetlab_welcome is not None:
            params['com_mynetlab_welcome'] = com_mynetlab_welcome

        return self.call('user.community.add', **params)

    def user_community_remove(self, *, com_id):
        """
        This method allows you to remove a community.

        :param int com_id: Community identifier.

        :returns: Success or Failure
        """

        return self.call('user.community.remove', com_id=com_id)

    def user_community_find(self, *, com_full_name, **kwargs):
        """
        This method allows you to find a community based on ``com_full_name``.

        :param str com_full_name: Community full name.

        :returns: The Community ID on success.
        :rtype: int
        """

        method = "user.community.find"
        return int(self.call(method, com_full_name=com_full_name, **kwargs))

    def user_community_get(self, *, com_id, properties="default", **kwargs):
        """
        This method allows you to retrieve a community and its properties.

        :param int com_id: Community ID.
        :param properties: Properties to return. Accepts string or list.
                        See below for all available properties. You may also use **"all"** to include all properties.
        :type properties: list or str

        :returns: A dictionary with the properties requested

        **Properties**

        com_alt_banner
            Alternate community banner.
        com_alt_banner_height
            Alternate banner height.
        com_alt_banner_width
            Alternate banner width.
        com_enabled
            Is community enabled.
        com_full_name
            Community full name.
        com_id
            Community identifier.
        com_max_slots_per_res
            Max amount of slots per reservation.
        com_min_hours_btw_res
            Minimum hours between reservations.
        com_mynetlab_news
            Community MyNETLAB+ news entry.
        com_mynetlab_welcome
            Community MyNETLAB+ welcome message.
        """

        method = "user.community.get"
        if properties == 'all':
            properties = self.community_props
        if properties == 'default':
            properties = self.community_default_props

        return self.call(method, com_id=com_id, properties=properties, **kwargs)

    def user_community_list(self, *, properties="default", sort_property="com_id", **kwargs):
        """
        Retrieve a list of communities.

        :param properties: Properties to return. Accepts string or list.
                        See below for all available properties. You may also use **"all"** to include all properties.
        :type properties: list or str
        :param str sort_property: Property to sort list by. Only ``com_id`` (default) and ``com_full_name`` are
                   valid options.

        :returns: A dictionary with the properties requested

        **Properties**

        com_alt_banner
            Alternate community banner.
        com_alt_banner_height
            Alternate banner height.
        com_alt_banner_width
            Alternate banner width.
        com_enabled
            Is community enabled.
        com_full_name
            Community full name.
        com_id
            Community identifier.
        com_max_slots_per_res
            Max amount of slots per reservation.
        com_min_hours_btw_res
            Minimum hours between reservations.
        com_mynetlab_news
            Community MyNETLAB+ news entry.
        com_mynetlab_welcome
            Community MyNETLAB+ welcome message.
        """

        method = "user.community.list"
        if properties == 'all':
            properties = self.community_props
        if properties == 'default':
            properties = self.community_default_props

        return self.call(method, properties=properties, sort_property=sort_property, **kwargs)

    def user_logins_system_get(self, **kwargs):
        """
        This method allows you to query the User Logins status and number of active logins.

        :returns: A dictionary with the following properties

        active
            Number of active users logged in.
        enabled
            User logins status.
        """
        method = "user.logins.system.get"
        return self.call(method, **kwargs)

    def user_logins_system_set(self, *, enabled, logout=False, **kwargs):
        """
        This method allows you to enable and disable user logins.

        :param bool enabled: Enable or disable user logins.
        :param bool logout: Logout users that are already logged in.

        :returns: A dictionary with the following properties

        changed
            Whether user login status was changed.
        logouts
            Number of users logged out.
        """
        method = "user.logins.system.set"
        return self.call(method, enabled=enabled, logout=logout, **kwargs)



NL_OK = "OK"
NL_PASSED = "PASSED"
NL_ERROR = "ERROR"
HDR_SEVERITY_LEVELS = {
    'FATAL': 1,
    'ERROR': 2,
    'WARN': 3,
    'SUCCESS': 4,
    'INFO': 5,
    'TRACE0': 6,
    'TRACE1': 7,
    'TRACE2': 8
}
